/* kbstate.c - detect keyboard states from shell
 * jamesbond 2013, 2015
 * re-written, ideas originally from http://stackoverflow.com/questions/3649874/how-to-get-keyboard-state-in-linux
 * License: Public Domain
 */
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/input.h>
#include <stdio.h>
#include <limits.h>

#define BITMASK(b) (1 << ((b) % CHAR_BIT))
#define BITSLOT(b) ((b) / CHAR_BIT)
#define BITSET(a, b) ((a)[BITSLOT(b)] |= BITMASK(b))
#define BITCLEAR(a, b) ((a)[BITSLOT(b)] &= ~BITMASK(b))
#define BITTEST(a, b) ((a)[BITSLOT(b)] & BITMASK(b))
#define BITNSLOTS(nb) ((nb + CHAR_BIT - 1) / CHAR_BIT)

//#define DEBUG
#ifdef DEBUG
	#define D if(1)
#else
	#define D if(0)
#endif

struct keycode_and_name {
	int code;
	char* name;
};

struct keycode_and_name keys[] = {
	{KEY_LEFTSHIFT, "lshift"},
	{KEY_LEFTALT, "lalt"},
	{KEY_LEFTCTRL, "lctrl"},
	{KEY_RIGHTSHIFT, "rshift"},
	{KEY_RIGHTALT, "ralt"},
	{KEY_RIGHTCTRL, "rctrl"},
	
	{KEY_CAPSLOCK, "capslock"},
	{KEY_NUMLOCK, "numlock"},
	{KEY_SCROLLLOCK, "scrolllock"},
	{KEY_LEFTMETA, "lmeta"},
	{KEY_RIGHTMETA, "rmeta"},
	{KEY_COMPOSE, "compose"},
	{0, 0}
};

void show_key_pressed(char *outmap) {
	int numkeys=0;
	struct keycode_and_name *p = keys;
	while (p->code) {
		if (BITTEST(outmap,p->code)) {
			printf("%s ",p->name);
			numkeys++;
		}
		p++;
	}	
	if (numkeys>0) printf("\n");
}

void check_key_pressed(char *inmap, char *outmap) {
	int  *key;
	struct keycode_and_name *p = keys;
	while (p->code) {
		if (BITTEST(inmap, p->code)) BITSET(outmap,p->code);
		p++;
	}
}

// put these in BSS to conserve stack
char path[256];
char inmap[KEY_MAX/CHAR_BIT + 1];  // Byte array the size of the number of keys
char outmap[KEY_MAX/CHAR_BIT + 1]; 

/*********** main ***********/
int main ( int argc, char *argv[], char *env[] )
{
	struct stat sb;
	int devno;
	int devfd;
	int ok;
	
	// look at all /dev/input/event* devices
	devno=0;
	memset(outmap, 0, sizeof(outmap));
	while(1) {
		snprintf(path,sizeof(path),"/dev/input/event%d",devno);
		if (stat(path, &sb)) break;
		D printf("Found %s ",path);
		devno++;
		
		ok = 0;
		if ( (devfd = open (path, O_RDONLY)) >= 0) {
			memset(inmap, 0, sizeof(inmap)); // clean map for next device
			if (ioctl(devfd, EVIOCGKEY(sizeof(inmap)), inmap) >= 0) {
				ok = 1;
				check_key_pressed (inmap, outmap);
			}
			close(devfd);
		}
		D printf("%s working\n",ok?"":"not ");
	}
	show_key_pressed(outmap);
	return 0;
}
